/* 
 *
 *  Copyright 1996 Claude Lecommandeur. See COPYRIGHT file for details
 *
 *  Author:  Claude Lecommandeur [ Claude.Lecommandeur@epfl.ch ][ August 1996 ]
 *
 */

#include "config.h"
#include <errno.h>
#include <sys/types.h>

#if defined(HAVE_SYS_TIME_H)
#   include <sys/time.h>
#endif

#if defined(HAVE_TIME_H)
#   include <time.h>
#endif

#if defined(HAVE_SYS_SELECT_H)
#include <sys/select.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xproto.h>
static int errorHandler(Display *, XErrorEvent *);
#include <X.h>

static void initevents ();
static void initcursors ();
static void initcolors ();
static void initGCs ();

Display	*dpy;  /* only one display at a time (for now) */
XFont	Xfonts [MAXFONTS];
XGC	Xgcs [MAXGCS];
int	nfonts		= 0;
Bool	debug		= False;
Bool	lockdebug	= False;
Bool	busydisplay	= False;
int	cursorFont;
long	colors[256];

static int errorHandler(Display *, XErrorEvent *);
static int (*previousErrorHandler)(Display *, XErrorEvent *);
extern int isOnGrabList(Window);

void kaffe_awt_simple_NativeDisplay_openDisplay (NativeDisplay display) {
    char	*displayname;
    double	xres, yres;
    int		scr;

    displayname = makeCString (unhand(display)->name);
    if (debug) fprintf (stderr, "openDisplay : display = %s\n", displayname);

    dpy = XOpenDisplay (displayname);
    if (dpy == (Display*) 0) {
	fprintf (stderr, "Can't open display : %s\n", displayname);
	exit (1);
    }
    scr = DefaultScreen (dpy);
    xres = ((((double) DisplayWidth (dpy, scr)) * 25.4) / ((double) DisplayWidthMM(dpy, scr)));
    yres = ((((double) DisplayHeight(dpy, scr)) * 25.4) / ((double) DisplayHeightMM(dpy, scr)));

    previousErrorHandler = XSetErrorHandler(errorHandler);
    unhand(display)->width	= DisplayWidth  (dpy, scr);
    unhand(display)->height	= DisplayHeight (dpy, scr);
    unhand(display)->resolution	= (int) xres;
    initevents();
    initcursors();
    initcolors();
    initGCs ();
}

void kaffe_awt_simple_NativeDisplay_syncDisplay (NativeDisplay display) {
    if (debug) fprintf (stderr, "syncDisplay\n");
    XSync (dpy, 0);
}

void kaffe_awt_simple_NativeDisplay_closeDisplay (NativeDisplay display) {
    if (debug) fprintf (stderr, "closeDisplay\n");
    LOCKDISPLAY ("closeDisplay");
    XCloseDisplay (dpy);
    RELEASEDISPLAY ("closeDisplay");
}

static int java_events[LASTEvent];
static int java_actionkeys[256];

static void initevents () {
    int i;
    for (i = 0; i < LASTEvent; i++) { java_events [i] = 0; }
    java_events[ButtonPress]	= kaffe_awt_simple_NativeEvent_MOUSE_DOWN;
    java_events[ButtonRelease]	= kaffe_awt_simple_NativeEvent_MOUSE_UP;
    java_events[KeyPress]	= kaffe_awt_simple_NativeEvent_KEY_PRESS;
    java_events[KeyRelease]	= kaffe_awt_simple_NativeEvent_KEY_RELEASE;
    java_events[EnterNotify]	= kaffe_awt_simple_NativeEvent_MOUSE_ENTER;
    java_events[LeaveNotify]	= kaffe_awt_simple_NativeEvent_MOUSE_EXIT;
    java_events[MotionNotify]	= kaffe_awt_simple_NativeEvent_MOUSE_DRAG;
    java_events[Expose]		= kaffe_awt_simple_NativeEvent_WINDOW_EXPOSE;
    java_events[FocusIn]	= kaffe_awt_simple_NativeEvent_GOT_FOCUS;
    java_events[FocusOut]	= kaffe_awt_simple_NativeEvent_LOST_FOCUS;
    java_events[ClientMessage]  = kaffe_awt_simple_NativeEvent_WINDOW_DESTROY;
    java_events[ConfigureNotify]= kaffe_awt_simple_NativeEvent_WINDOW_MOVED;
    
    for (i = 0; i < 256; i++) { java_actionkeys[i] = -1; }
    java_actionkeys [XK_Home  & 0x00FF]	= kaffe_awt_simple_NativeEvent_HOME;
    java_actionkeys [XK_Left  & 0x00FF]	= kaffe_awt_simple_NativeEvent_LEFT;
    java_actionkeys [XK_Up    & 0x00FF]	= kaffe_awt_simple_NativeEvent_UP;
    java_actionkeys [XK_Right & 0x00FF]	= kaffe_awt_simple_NativeEvent_RIGHT;
    java_actionkeys [XK_Down  & 0x00FF]	= kaffe_awt_simple_NativeEvent_DOWN;
    java_actionkeys [XK_Prior & 0x00FF]	= kaffe_awt_simple_NativeEvent_PGUP;
    java_actionkeys [XK_Next  & 0x00FF]	= kaffe_awt_simple_NativeEvent_PGDN;
    java_actionkeys [XK_End   & 0x00FF]	= kaffe_awt_simple_NativeEvent_END;
    java_actionkeys [XK_F1    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F1;
    java_actionkeys [XK_F2    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F2;
    java_actionkeys [XK_F3    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F3;
    java_actionkeys [XK_F4    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F4;
    java_actionkeys [XK_F5    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F5;
    java_actionkeys [XK_F6    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F6;
    java_actionkeys [XK_F7    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F7;
    java_actionkeys [XK_F8    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F8;
    java_actionkeys [XK_F9    & 0x00FF]	= kaffe_awt_simple_NativeEvent_F9;
    java_actionkeys [XK_F10   & 0x00FF]	= kaffe_awt_simple_NativeEvent_F10;
    java_actionkeys [XK_F11   & 0x00FF]	= kaffe_awt_simple_NativeEvent_F11;
    java_actionkeys [XK_F12   & 0x00FF]	= kaffe_awt_simple_NativeEvent_F12;
}

int getJavaMask (int xMask) {
    int ret = 0;

    if (xMask & Mod1Mask)	ret |= kaffe_awt_simple_NativeEvent_ALT_MASK;
    if (xMask & ControlMask)	ret |= kaffe_awt_simple_NativeEvent_CTRL_MASK;
    if (xMask & ShiftMask)	ret |= kaffe_awt_simple_NativeEvent_SHIFT_MASK;
    if (xMask & Mod2Mask)	ret |= kaffe_awt_simple_NativeEvent_META_MASK;
}

static void initcursors () {
    LOCKDISPLAY ("initcursors");
    cursorFont = loadFont ("cursor");
    RELEASEDISPLAY ("initcursors");
}

static void initcolors () {
    int i;
    for (i = 0; i < 256; i++) {
	colors [i] = 0;
    }
}

static int maskwidth(unsigned long mask)
{
    int bit = 0;

    while ((mask & 1) == 1) {
	bit++;
	mask >>= 1;
    }
    return bit;
}

static int maskshift(unsigned long mask)
{
    int bit = 0;

    if (mask == 0) return 0;

    while ((mask & 1) == 0) {
	bit++;
	mask >>= 1;
    }
    return bit;
}

static int makergbpixel(Visual* visual, int rgb)
{
    int rs, rw;
    int gs, gw;
    int bs, bw;
    int r, g, b;

    rs = maskshift(visual->red_mask);
    rw = maskwidth(visual->red_mask >> rs);
    gs = maskshift(visual->green_mask);
    gw = maskwidth(visual->green_mask >> gs);
    bs = maskshift(visual->blue_mask);
    bw = maskwidth(visual->blue_mask >> bs);

    r = (rgb >> 16) & 0xff;
    g = (rgb >> 8) & 0xff;
    b = rgb & 0xff;

    return ((r >> (8 - rw)) << rs)
	| ((g >> (8 - gw)) << gs)
	| ((b >> (8 - bw)) << bs);
}

int getcolor (long rgb) {
    int		i;
    XColor	xcolor;
    Colormap	cmap = DefaultColormap (dpy, 0);
    Visual*	visual = DefaultVisual (dpy, 0);

    if (visual->class == TrueColor) {
	return makergbpixel(visual, rgb);
    }
    for (i = 0; i < 256; i++) {
	if (colors [i] == rgb) return i;
    }
    xcolor.red   = ((rgb >> 8) & 0xFF00) | ((rgb >> 16) & 0x00FF);
    xcolor.green = ((rgb >> 0) & 0xFF00) | ((rgb >>  8) & 0x00FF);
    xcolor.blue  = ((rgb << 8) & 0xFF00) | ((rgb >>  0) & 0x00FF);
    xcolor.pixel = 0;

    LOCKDISPLAY ("getcolor:XAllocColor");
    if (!XAllocColor (dpy, cmap, &xcolor)) {
	fprintf (stderr, "XAllocColor failed\n");
    }
    RELEASEDISPLAY ("getcolor:XAllocColor");
    colors [xcolor.pixel] = rgb;
    return xcolor.pixel;
}

static void initGCs () {
    int i;
    for (i = 0; i < MAXGCS; i++) {
	Xgcs [i] = (GC)0;
    }
}

int allocateGC (Window win, int iid) {
    XGCValues	values;
    int		i;

    for (i = 0; i < MAXGCS; i++) {
	if (Xgcs [i] == (GC)0) break;
    }
    if (i == MAXGCS) {
	fprintf (stderr, "No more native GC available, exiting...\n");
	exit (1);
    }
    values.font	= Xfonts[iid]->fid;
    LOCKDISPLAY ("allocateGC:createGC");
    Xgcs [i] = XCreateGC (dpy, win, GCFont, &values);
    RELEASEDISPLAY ("allocateGC:createGC");
    return i;
}

void freeGC (int gid) {
    LOCKDISPLAY ("freeGC:freeGC");
    XFreeGC (dpy, Xgcs [gid]);
    RELEASEDISPLAY ("freeGC:freeGC");
    Xgcs [gid] = (GC)0;
}

int getcolorrgb (unsigned short red, unsigned short green, unsigned short blue) {
    int		rgb;

    red   = red   >> 8;
    green = green >> 8;
    blue  = blue  >> 8;
    rgb   = (red << 16) | (green << 8) | (blue << 0);
    return getcolor (rgb);
}

jint kaffe_awt_simple_NativeDisplay_nextEvent (NativeDisplay display, NativeEvent event) {
    XEvent	xevent;
    int		found;
    fd_set	mask;
    int		fd;
    struct timeval timeout;
    static Time	lastclick  = 0;
    static int	clickcount = 1;
    int		len;
    char	key [16];

    timeout.tv_sec  = 0;
    timeout.tv_usec = 100000;

    if (debug) fprintf (stderr, "nextEvent called\n");
    if (!XEventsQueued (dpy, QueuedAfterFlush)) {
	fd = ConnectionNumber (dpy);
	FD_ZERO (&mask);
	FD_SET (fd, &mask);
	found = select (fd + 1, &mask, (fd_set*) 0, (fd_set*) 0, &timeout);
	if (found <= 0) return False;
    }
    XNextEvent (dpy, &xevent);

/* To allow for a modal cascade, all events other than Expose need to be
   checked in case they are for a window that should be disabled by an
   active modal dialog.
 */   
    if (xevent.type != Expose && !isOnTopDialog(xevent.xany.window))
	return False;
    
    if (debug) fprintf (stderr, "nextEvent:gotevent: type = %d\n", xevent.type);
    switch (xevent.type) {
	case ButtonPress :
	    if ((xevent.xbutton.time - lastclick) < DoubleClickTime) clickcount++;
	    else clickcount = 1;

	case ButtonRelease :
	    if (xevent.type == ButtonRelease) lastclick = xevent.xbutton.time;

	    unhand(event)->clickCount	= clickcount;
	    unhand(event)->id		= java_events [xevent.type];
	    unhand(event)->modifiers	= getJavaMask (xevent.xbutton.state);
	    unhand(event)->when		= xevent.xbutton.time;
	    unhand(event)->x		= xevent.xbutton.x;
	    unhand(event)->y		= xevent.xbutton.y;
	    break;

	case KeyPress :
	case KeyRelease :
	    len = XLookupString (&(xevent.xkey), key, 16, NULL, NULL);
	    if (len != 1) {
		KeySym keysym;
		keysym = XLookupKeysym ((XKeyEvent*) &xevent, 0);
		if (! keysym) return False;

		if (xevent.type == KeyPress) 
		    unhand(event)->id	= kaffe_awt_simple_NativeEvent_KEY_ACTION;
		else
		    unhand(event)->id	= kaffe_awt_simple_NativeEvent_KEY_ACTION_RELEASE;
		unhand(event)->key	= java_actionkeys[keysym&0x00FF];
	    } else {
		unhand(event)->id	= java_events [xevent.type];
		unhand(event)->key	= key [0];
	    }
	    unhand(event)->modifiers	= getJavaMask (xevent.xkey.state);
	    unhand(event)->when		= xevent.xkey.time;
	    unhand(event)->x		= xevent.xkey.x;
	    unhand(event)->y		= xevent.xkey.y;
	    break;

	case EnterNotify :
	case LeaveNotify :
	    unhand(event)->id		= java_events [xevent.type];
	    unhand(event)->modifiers	= getJavaMask (xevent.xcrossing.state);
	    unhand(event)->when		= xevent.xcrossing.time;
	    unhand(event)->x		= xevent.xcrossing.x;
	    unhand(event)->y		= xevent.xcrossing.y;
	    break;

	case MotionNotify :
	    unhand(event)->id		= java_events [xevent.type];
	    unhand(event)->modifiers	= getJavaMask (xevent.xmotion.state);
	    unhand(event)->when		= xevent.xmotion.time;
	    unhand(event)->x		= xevent.xmotion.x;
	    unhand(event)->y		= xevent.xmotion.y;
	    break;

	case Expose :
	    unhand(event)->id		= java_events [xevent.type];
	    unhand(event)->x		= xevent.xexpose.x;
	    unhand(event)->y		= xevent.xexpose.y;
	    break;

	case FocusIn :
	case FocusOut :
	    unhand(event)->id		= java_events [xevent.type];
	    break;

        case ClientMessage:
/* We are only interested in WM_DELETE_WINDOW messages */
	    if (xevent.xclient.message_type == XInternAtom(dpy, "WM_PROTOCOLS", False)
		&& xevent.xclient.data.s[0] == XInternAtom(dpy, "WM_DELETE_WINDOW", False)) {
		
		unhand(event)->id = java_events [xevent.type];
	    }
	    break;

/* ConfigureNotify just sends WINDOW_MOVED messages for the mo.
   There does not seem to be a mechanism for reporting size changes */
        case ConfigureNotify:
	    unhand(event)->id = java_events [xevent.type];
	    unhand(event)->x  = xevent.xconfigure.x;
	    unhand(event)->y  = xevent.xconfigure.y;
	    break;
	    
    default :
	    return (False);
    }
    unhand(event)->winid = xevent.xany.window;
    return True;
}

/* Deal with any X protocol errors here.
   Small hack factor. I may deal with the problem properly later...
   When a dialog is closed there may sill be expose events in the event
   queue for it's children. But by the time we process those events the dialog
   and it's children have been destroyed.

   This bit simply ignores any errors caused by events for children
   All other errors are propcessed by the previous error handler
  */
static int errorHandler(Display *dpy, XErrorEvent *xee) {

    if (xee->error_code == BadWindow) return 0;
     return previousErrorHandler(dpy, xee);
}

#if !defined(HAVE_USLEEP)
int usleep (unsigned long usec) {
    struct timeval timeout;
    timeout.tv_usec = usec % (unsigned long) 1000000;
    timeout.tv_sec = usec / (unsigned long) 1000000;
    select(0, (void *) 0, (void *) 0, (void *) 0, &timeout);
    return 0;
}
#endif
