/* 
 *
 *  Copyright 1996 Claude Lecommandeur. See COPYRIGHT file for details
 *
 *  Author:  Claude Lecommandeur [ Claude.Lecommandeur@epfl.ch ][ August 1996 ]
 *
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X.h>

jint kaffe_awt_simple_NativeDisplay_loadFont (NativeDisplay display,
				JavaString name, jint style, jint size) {
    char fullname[256];
    char *fontname;
    char cfsize[4];
    int ifont;

    fontname = makeCString (name);
    strcpy (fullname, "-adobe-");
    if (!strcmp(fontname, "Helvetica"))  { strcat (fullname, "helvetica"); } else
    if (!strcmp(fontname, "TimesRoman")) { strcat (fullname, "times");     } else
    if (!strcmp(fontname, "Courier"))    { strcat (fullname, "courier");   }
    else strcat (fullname, "courier");

    if (style == java_awt_Font_BOLD) { strcat (fullname, "-bold-"); }
    else strcat (fullname, "-medium-");

    if (style == java_awt_Font_ITALIC) { strcat (fullname, "i"); }
    else strcat (fullname, "r");

    strcat (fullname, "-normal--");
    sprintf (cfsize, "%d", size);
    strcat (fullname, cfsize);
    strcat (fullname, "-*-*-*-*-*-iso8859-1");
    ifont = loadFont (fullname);
    return ((jint)ifont);
}

int loadFont (char *name) {
    XFont xfont;
    XFontStruct *fontstruct;

    if (debug) fprintf (stderr, "loadFont:name = %s\n", name);
    LOCKDISPLAY ("loadFont");
    fontstruct = XLoadQueryFont (dpy, name);
    RELEASEDISPLAY ("loadFont");
    if (!fontstruct) {
	fprintf (stderr, "unable to open font %s\n", name);
	fontstruct = XLoadQueryFont (dpy, "fixed");
	if (!fontstruct) {
	    fprintf (stderr, "unable to open fixed font, exiting...\n");
	    exit(1);
	}
    }
    if (debug) fprintf (stderr, "loadFont:fontid = %x\n", fontstruct->fid);
    xfont = (struct _XFont*) malloc (sizeof (struct _XFont));
    xfont->iid = nfonts;
    xfont->fid = fontstruct->fid;
    xfont->fontstruct = fontstruct;
    
    Xfonts [nfonts] = xfont;
    nfonts++;
    return (nfonts - 1);
}

void kaffe_awt_simple_NativeDisplay_getFontMetrics (NativeDisplay display, jlong fontid,
				NativeFontMetrics metrics)
{
    int	iid;
    int i;
    XFontStruct *fontstruct;
    HArrayOfInt *widths;

    if (debug) fprintf (stderr, "getFontMetrics\n");
    fontstruct = Xfonts [(int)fontid]->fontstruct;

    unhand(metrics)->ascent  = fontstruct->ascent;
    unhand(metrics)->descent = fontstruct->descent;
    unhand(metrics)->leading = 0;
    unhand(metrics)->maxascent  = fontstruct->min_bounds.ascent;
    unhand(metrics)->maxdescent = fontstruct->min_bounds.descent;

    if (fontstruct->min_char_or_byte2 > 256) fontstruct->min_char_or_byte2 = 256;
    if (fontstruct->max_char_or_byte2 > 256) fontstruct->max_char_or_byte2 = 256;
    widths = unhand(metrics)->widths;
    for (i = fontstruct->min_char_or_byte2; i < fontstruct->max_char_or_byte2; i++) {
	(unhand(widths)->body) [i] = fontstruct->per_char [i].width;
    }

}

jint kaffe_awt_simple_NativeDisplay_bytesWidth (NativeDisplay display,
		NativeFontMetrics metrics, HArrayOfByte *data, jint off, jint len)
{
    jint	ret;
    int		iid;
    char	*cdata;
    XFontStruct *fontstruct;

    if (debug) fprintf (stderr, "bytesWidth\n");
    iid = unhand(metrics)->fid;
    fontstruct = Xfonts[iid]->fontstruct;
    cdata = unhand(data)->body;

    ret = XTextWidth (fontstruct, cdata + off, len);
    if (debug) fprintf (stderr, "ret = %d\n", ret);
    return (ret);
}

jint kaffe_awt_simple_NativeDisplay_charsWidth (NativeDisplay display,
		NativeFontMetrics metrics, HArrayOfChar *data, jint off, jint len)
{
    if (debug) fprintf (stderr, "charsWidth\n");
    return 100;
}

jint kaffe_awt_simple_NativeDisplay_charWidth (NativeDisplay display,
#ifdef KAFFE
		NativeFontMetrics metrics, jint ch) /* looks like a bug in kaffeh */
#else
		NativeFontMetrics metrics, unicode ch)
#endif
{
    if (debug) fprintf (stderr, "charWidth\n");
    return 10;
}

jint kaffe_awt_simple_NativeDisplay_stringWidth (NativeDisplay display,
		NativeFontMetrics metrics, JavaString jstr)
{
    jint	ret;
    int		iid;
    char	*cstr;
    XFontStruct *fontstruct;

    if (debug) fprintf (stderr, "stringWidth, jstr = %x\n", jstr);
    cstr = makeCString (jstr);
    if (debug) fprintf (stderr, "cstr = %s\n", cstr);
    iid = unhand(metrics)->fid;
    fontstruct = Xfonts[iid]->fontstruct;
    ret = XTextWidth (fontstruct, cstr, strlen (cstr));
    if (debug) fprintf (stderr, "ret = %d\n", ret);
    return (ret);
}

