/* 
 *
 *  Copyright 1996 Claude Lecommandeur. See COPYRIGHT file for details
 *
 *  Author:  Claude Lecommandeur [ Claude.Lecommandeur@epfl.ch ][ August 1996 ]
 *
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X.h>

static Window	getWindow (NativeGraphics);

void kaffe_awt_simple_NativeDisplay_createGraphics (
			NativeDisplay display, NativeGraphics graphics) {
    Window	win;
    int		gid, iid;

    if (debug) fprintf (stderr, "createGraphics\n");
    win = getWindow (graphics);
    if (!win) return;

    iid	= unhand(graphics)->fontid;
    gid = allocateGC (win, iid);
    unhand(graphics)->gid = gid;
}

void kaffe_awt_simple_NativeDisplay_clearRect (
			NativeDisplay display, NativeGraphics graphics,
			jint x, jint y, jint width, jint height) {
    Window win;

    if (debug) fprintf (stderr, "clearRect\n");
    win = getWindow (graphics);
    LOCKDISPLAY ("clearRect");
    XClearArea (dpy, win, x, y, width, height, True);
    RELEASEDISPLAY ("clearRect");
}

void kaffe_awt_simple_NativeDisplay_copyArea (
			NativeDisplay display, NativeGraphics graphics,
			jint x, jint y, jint width, jint height, jint dx, jint dy) {
    Window	win;
    int		gid;

    if (debug) fprintf (stderr, "copyArea\n");
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;
    LOCKDISPLAY ("copyArea");
    XCopyArea (dpy, win, win, Xgcs[gid], x, y, width, height, dx, dy);
    RELEASEDISPLAY ("copyArea");
}

void kaffe_awt_simple_NativeDisplay_dispose (
				NativeDisplay display, NativeGraphics graphics)
{
    Window	win;
    int		gid;

    if (debug) fprintf (stderr, "dispose\n");
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;

    freeGC (gid);
    LOCKDISPLAY ("dispose");
    XDestroyWindow (dpy, win);
    RELEASEDISPLAY ("dispose");
}

void kaffe_awt_simple_NativeDisplay_drawArc (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y,
				jint width, jint height,
				jint startAngle, jint arcAngle)
{
    if (debug) fprintf (stderr, "drawArc\n");
}

jint kaffe_awt_simple_NativeDisplay_drawImage (
				NativeDisplay display, NativeGraphics graphics,
				JavaImage image,
				jint x, jint y, jint width, jint height,
				JavaColor color, JavaImageObserver observer)
{
    if (debug) fprintf (stderr, "drawImage\n");
}

void kaffe_awt_simple_NativeDisplay_drawLine (
				NativeDisplay display, NativeGraphics graphics,
				jint x1, jint y1, jint x2, jint y2)
{
    Window	win;
    int		gid;

    if (debug) fprintf (stderr, "drawLine : %d, %d, %d, %d\n", x1, y1, x2, y2);
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;
    LOCKDISPLAY ("drawLine");
    XDrawLine (dpy, win, Xgcs[gid], x1, y1, x2, y2);
    RELEASEDISPLAY ("drawLine");
}

void kaffe_awt_simple_NativeDisplay_drawOval (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height)
{
    if (debug) fprintf (stderr, "drawOval\n");
}

void kaffe_awt_simple_NativeDisplay_drawPolygon (
				NativeDisplay display, NativeGraphics graphics,
				HArrayOfInt *xPoints, HArrayOfInt *yPoints, jint nPoints)
{
    if (debug) fprintf (stderr, "drawPolygon\n");
}

void kaffe_awt_simple_NativeDisplay_drawRoundRect (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height,
				jint arcWidth, jint arcHeight)
{
    if (debug) fprintf (stderr, "drawRoundRect\n");
}

void kaffe_awt_simple_NativeDisplay_drawString (
				NativeDisplay display, NativeGraphics graphics,
				JavaString jstr, jint x, jint y)
{
    Window	win;
    int		gid;
    char	*cstr;

    cstr = makeCString (jstr);
    if (debug) fprintf (stderr, "drawString : str = %s\n", cstr);
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;
    LOCKDISPLAY ("drawString");
    XDrawString (dpy, win, Xgcs[gid], x, y, cstr, strlen (cstr));
    RELEASEDISPLAY ("drawString");
}

void kaffe_awt_simple_NativeDisplay_drawCursor (
				NativeDisplay display, NativeGraphics graphics,
				JavaString jname, jint x, jint y)
{
    Window	win;
    int		gid, iid;
    char	*cname;
    char	curs[2];
    XGCValues	values;

    cname = makeCString (jname);
    if (debug) fprintf (stderr, "drawCursor : name = %s\n", cname);
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;
    iid	= unhand(graphics)->fontid;

    curs [0] = 152;
    curs [1] = 0;
    LOCKDISPLAY ("drawCursor");
    values.font	= Xfonts[cursorFont]->fid;
    XChangeGC (dpy, Xgcs[gid], GCFont, &values);
    XDrawString (dpy, win, Xgcs[gid], x, y, curs, 1);
    values.font	= Xfonts[iid]->fid;
    XChangeGC (dpy, Xgcs[gid], GCFont, &values);
    RELEASEDISPLAY ("drawCursor");
}

void kaffe_awt_simple_NativeDisplay_fillArc (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height,
				jint startAngle, jint arcAngle)
{
    if (debug) fprintf (stderr, "fillArc\n");
}

void kaffe_awt_simple_NativeDisplay_fillOval (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height)
{
    if (debug) fprintf (stderr, "fillOval\n");
}

void kaffe_awt_simple_NativeDisplay_fillPolygon (
				NativeDisplay display, NativeGraphics graphics,
				HArrayOfInt *xPoints,
				HArrayOfInt *yPoints, jint nPoints)
{
    if (debug) fprintf (stderr, "fillPolygon\n");
}

void kaffe_awt_simple_NativeDisplay_fillRect (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height)
{
    Window	win;
    int		gid;

    if (debug) fprintf (stderr, "fillRect\n");
    win = getWindow (graphics);
    gid = unhand(graphics)->gid;
    LOCKDISPLAY ("fillRect");
    XFillRectangle (dpy, win, Xgcs[gid], x, y, width, height);
    RELEASEDISPLAY ("fillRect");
}

void kaffe_awt_simple_NativeDisplay_fillRoundRect (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y, jint width, jint height,
				jint arcWidth, jint arcHeight)
{
    if (debug) fprintf (stderr, "fillRoundRect\n");
}

void kaffe_awt_simple_NativeDisplay_setPaintMode (
				NativeDisplay display, NativeGraphics graphics)
{
    if (debug) fprintf (stderr, "setPaintMode\n");
}

void kaffe_awt_simple_NativeDisplay_setXORMode (
				NativeDisplay display, NativeGraphics graphics,
				JavaColor color)
{
    if (debug) fprintf (stderr, "setXORMode\n");
}

void kaffe_awt_simple_NativeDisplay_translate (
				NativeDisplay display, NativeGraphics graphics,
				jint x, jint y)
{
    if (debug) fprintf (stderr, "translate\n");
}

void kaffe_awt_simple_NativeDisplay_clipNone (NativeDisplay display, NativeGraphics graphics)
{
    int	gid;
    gid = unhand(graphics)->gid;
    XSetClipMask (dpy, Xgcs[gid], None);
}

void kaffe_awt_simple_NativeDisplay_setFont (NativeDisplay display, NativeGraphics graphics)
{
    int		gid;
    Font 	font;
    int  	iid;
    XGCValues	values;

    if (debug) fprintf (stderr, "setFont\n");
    gid = unhand(graphics)->gid;
    iid	= unhand(graphics)->fontid;
    values.font	= Xfonts[iid]->fid;

    LOCKDISPLAY ("setFont:XChangeGC");
    XChangeGC (dpy, Xgcs[gid], GCFont, &values);
    RELEASEDISPLAY ("setFont:XChangeGC");
}

void kaffe_awt_simple_NativeDisplay_setColor (NativeDisplay display, NativeGraphics graphics)
{
    int		gid;
    long	rgb;
    XGCValues	values;
    int		color;

    if (debug) fprintf (stderr, "setColor\n");
    gid = unhand(graphics)->gid;
    rgb	= unhand(graphics)->rgb;
    values.foreground = getcolor (rgb);

    LOCKDISPLAY ("setColor:XChangeGC");
    XChangeGC (dpy, Xgcs[gid], GCForeground, &values);
    RELEASEDISPLAY ("setColor:XChangeGC");
}

void kaffe_awt_simple_NativeDisplay_clipRect (NativeDisplay display, NativeGraphics graphics)
{
    int		gid;
    JavaRectangle clipRect;
    XRectangle	xrect;

    gid      = unhand(graphics)->gid;
    clipRect = unhand(graphics)->clipRect;
    if (clipRect) {
	xrect.x      = unhand(clipRect)->x;
	xrect.y      = unhand(clipRect)->y;
	xrect.width  = unhand(clipRect)->width;
	xrect.height = unhand(clipRect)->height;
	XSetClipRectangles (dpy, Xgcs[gid], 0, 0, &xrect, 1, Unsorted);
    }
}
/*
 *
 *         Utilities
 *
 */

static Window getWindow (NativeGraphics graphics) {
    NativeWindow	window;

    window = unhand(graphics)->window;
    if (!window) return (Window)0;
    return unhand(window)->id;
}

