/* 
 *
 *  Copyright 1996 Claude Lecommandeur. See COPYRIGHT file for details
 *
 *  Author:  Claude Lecommandeur [ Claude.Lecommandeur@epfl.ch ][ August 1996 ]
 *  Very heavily inspired (i.e. stolen) from NCSA Mosaic source.
 *
 */

#include <sys/types.h>
#include <stdio.h>
#include <setjmp.h>
#include "jpeglib.h"
#include <X11/Intrinsic.h>

#define MAXCOLORMAPSIZE         256

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;


METHODDEF void my_error_exit (j_common_ptr cinfo) {
    my_error_ptr myerr = (my_error_ptr) cinfo->err;
    fprintf (stderr, "Error reading JPEG image \n");
    longjmp (myerr->setjmp_buffer, 1);
}

unsigned char *ReadJPEG (char *filename, int *width, int *height, XColor *colors) {
    int scale = 65536 / (MAXCOLORMAPSIZE);
    int i;

    FILE			*infile;
    struct jpeg_decompress_struct	cinfo;
    struct my_error_mgr		jerr;
    unsigned char		*retBuffer = 0;
    unsigned char		*r;
    JSAMPARRAY			buffer;		/* Output row buffer */
    int				row_stride;	/* physical row width in output buffer */

    infile = fopen (filename, "r");
    if (!infile) {
	fprintf (stderr, "can't open file : filename\n");
	return 0;
    }
    cinfo.err = jpeg_std_error (&jerr.pub);
    jerr.pub.error_exit = my_error_exit;

    if (setjmp (jerr.setjmp_buffer)) {
	jpeg_destroy_decompress (&cinfo);
	fclose (infile);
	if (retBuffer) { free(retBuffer); }
	return 0;
    }
    jpeg_create_decompress (&cinfo);
    jpeg_stdio_src (&cinfo, infile);
    jpeg_read_header (&cinfo, TRUE);

    cinfo.quantize_colors		= TRUE; 
    cinfo.desired_number_of_colors	= 50;
    cinfo.two_pass_quantize		= TRUE;
    cinfo.out_color_space		= JCS_RGB;

    jpeg_start_decompress (&cinfo);

    if (!(retBuffer = (unsigned char *) malloc(cinfo.output_width 
			* cinfo.output_height * cinfo.output_components))) {
	jpeg_destroy_decompress( &cinfo);
	fprintf (stderr,"Couldn't create space for JPEG read\n");
	return (0);
    }
    row_stride = cinfo.output_width * cinfo.output_components;
    buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

    r = retBuffer;
    row_stride = cinfo.output_width * cinfo.output_components;
    while (cinfo.output_scanline < cinfo.output_height) {
	(void) jpeg_read_scanlines (&cinfo, buffer, 1);
	memcpy (r,buffer [0], row_stride);
	r += row_stride;
    }
    *width  = cinfo.output_width;
    *height = cinfo.output_height;

    if (cinfo.out_color_components  == 3) {
	for (i = 0; i < cinfo.actual_number_of_colors; i++) {
	    colors [i].red	= cinfo.colormap [0][i] * scale;
	    colors [i].green	= cinfo.colormap [1][i] * scale;
	    colors [i].blue	= cinfo.colormap [2][i] * scale;
	    colors [i].pixel	= i;
	    colors [i].flags	= DoRed | DoGreen | DoBlue;
	}
    } else {
	for (i=0; i < cinfo.actual_number_of_colors; i++) {
	    colors [i].red =
		colors [i].green =
		colors [i].blue  = cinfo.colormap [0][i] * scale;
	    colors [i].pixel = i;
	    colors [i].flags = DoRed | DoGreen | DoBlue;
	}
    }
    (void) jpeg_finish_decompress (&cinfo);
    jpeg_destroy_decompress (&cinfo);
    return retBuffer;
}
